import React, { createContext, useContext, useRef, useReducer, useState, useEffect, useLayoutEffect } from 'react';
import { compose, createStore as createStore$1, applyMiddleware } from 'redux';
import reduxThunk from 'redux-thunk';
import { isDraft, current, Immer } from 'immer';

const StoreContext = createContext();

// To get around it, we can conditionally useEffect on the server (no-op) and
// useLayoutEffect in the browser. We need useLayoutEffect to ensure the store
// subscription callback always has the selector from the latest render commit
// available, otherwise a store update may happen between render and the effect,
// which may cause missed updates; we also must ensure the store subscription
// is created synchronously, otherwise a store update may occur before the
// subscription is created and an inconsistent state may be observed

const useIsomorphicLayoutEffect = typeof window !== 'undefined' ? useLayoutEffect : useEffect;
function createStoreStateHook(Context) {
  return function useStoreState(mapState, equalityFn) {
    const store = useContext(Context);
    const mapStateRef = useRef(mapState);
    const stateRef = useRef();
    const mountedRef = useRef(true);
    const subscriptionMapStateError = useRef();
    const [, forceRender] = useReducer(s => s + 1, 0);

    if (subscriptionMapStateError.current || mapStateRef.current !== mapState || stateRef.current === undefined) {
      try {
        stateRef.current = mapState(store.getState());
      } catch (err) {
        if (process.env.NODE_ENV === 'development') {
          let errorMessage = `Error in useStoreState: ${err.message}.`;

          if (subscriptionMapStateError.current) {
            errorMessage += `\nMaybe related to:\n${subscriptionMapStateError.current.stack}`;
          }

          throw new Error(errorMessage);
        }

        throw subscriptionMapStateError.current || err;
      }
    }

    useIsomorphicLayoutEffect(() => {
      mapStateRef.current = mapState;
      subscriptionMapStateError.current = undefined;
    });
    useIsomorphicLayoutEffect(() => {
      const checkMapState = () => {
        try {
          const newState = mapStateRef.current(store.getState());
          const isStateEqual = typeof equalityFn === 'function' ? equalityFn(stateRef.current, newState) : stateRef.current === newState;

          if (isStateEqual) {
            return;
          }

          stateRef.current = newState;
        } catch (err) {
          // see https://github.com/reduxjs/react-redux/issues/1179
          // There is a possibility mapState will fail due to stale state or
          // props, therefore we will just track the error and force our
          // component to update. It should then receive the updated state
          subscriptionMapStateError.current = err;
        }

        if (mountedRef.current) {
          forceRender({});
        }
      };

      const unsubscribe = store.subscribe(checkMapState);
      checkMapState();
      return () => {
        mountedRef.current = false;
        unsubscribe();
      };
    }, []);
    return stateRef.current;
  };
}
const useStoreState = createStoreStateHook(StoreContext);
function createStoreActionsHook(Context) {
  return function useStoreActions(mapActions) {
    const store = useContext(Context);
    return mapActions(store.getActions());
  };
}
const useStoreActions = createStoreActionsHook(StoreContext);
function createStoreDispatchHook(Context) {
  return function useStoreDispatch() {
    const store = useContext(Context);
    return store.dispatch;
  };
}
const useStoreDispatch = createStoreDispatchHook(StoreContext);
function useStore() {
  return useContext(StoreContext);
}
function createStoreRehydratedHook(Context) {
  return function useStoreRehydrated() {
    const store = useContext(Context);
    const [rehydrated, setRehydrated] = useState(false);
    useEffect(() => {
      store.persist.resolveRehydration().then(() => setRehydrated(true));
    }, []);
    return rehydrated;
  };
}
const useStoreRehydrated = createStoreRehydratedHook(StoreContext);
function createTypedHooks() {
  return {
    useStoreActions,
    useStoreDispatch,
    useStoreState,
    useStoreRehydrated,
    useStore
  };
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const actionSymbol = '$ep_a';
const actionOnSymbol = '$ep_aO';
const computedSymbol = '$ep_c';
const effectOnSymbol = '$ep_e';
const persistSymbol = '$ep_p';
const reducerSymbol = '$ep_r';
const thunkOnSymbol = '$ep_tO';
const thunkSymbol = '$ep_t';

const debug = state => {
  if (isDraft(state)) {
    return current(state);
  }

  return state;
}; // eslint-disable-next-line no-unused-vars

const memo = (fn, cacheSize) => {
  if (process.env.NODE_ENV === 'development') {
    // eslint-disable-next-line no-console
    console.log('Easy Peasy no longer ships with a memoization util. We suggest installing something like fast-memoize as an alternative.');
  }

  return fn;
};
const actionOn = (targetResolver, fn) => {
  return {
    [actionOnSymbol]: true,
    fn,
    targetResolver
  };
};
const action = fn => {
  return {
    [actionSymbol]: true,
    fn
  };
};
const defaultStateResolvers = [state => state];
const computed = (fnOrStateResolvers, fn) => {
  if (typeof fn === 'function') {
    return {
      [computedSymbol]: true,
      fn,
      stateResolvers: fnOrStateResolvers
    };
  }

  return {
    [computedSymbol]: true,
    fn: fnOrStateResolvers,
    stateResolvers: defaultStateResolvers
  };
};
function unstable_effectOn(dependencyResolvers, fn) {
  return {
    [effectOnSymbol]: true,
    dependencyResolvers,
    fn
  };
}
function generic(value) {
  return value;
}
const persist = (model, config) => {
  // if we are not running in a browser context this becomes a no-op
  return typeof window === 'undefined' ? model : _extends({}, model, {
    [persistSymbol]: config
  });
};
const thunkOn = (targetResolver, fn) => {
  return {
    [thunkOnSymbol]: true,
    fn,
    targetResolver
  };
};
const thunk = fn => {
  return {
    [thunkSymbol]: true,
    fn
  };
};
const reducer = fn => {
  return {
    [reducerSymbol]: true,
    fn
  };
};

/**
 * We create our own immer instance to avoid potential issues with autoFreeze
 * becoming default enabled everywhere. We want to disable autofreeze as it
 * does not suit the design of Easy Peasy.
 * https://github.com/immerjs/immer/issues/681#issuecomment-705581111
 */

const easyPeasyImmer = new Immer({
  useProxies: true,
  autoFreeze: false
});
/*!
 * is-plain-object <https://github.com/jonschlinkert/is-plain-object>
 *
 * Copyright (c) 2014-2017, Jon Schlinkert.
 * Released under the MIT License.
 */

function isObject(o) {
  return Object.prototype.toString.call(o) === '[object Object]';
}

function isPlainObject(o) {
  if (isObject(o) === false) return false; // If has modified constructor

  const ctor = o.constructor;
  if (ctor === undefined) return true; // If has modified prototype

  const prot = ctor.prototype;
  if (isObject(prot) === false) return false; // If constructor does not have an Object-specific method
  // eslint-disable-next-line no-prototype-builtins

  if (prot.hasOwnProperty('isPrototypeOf') === false) {
    return false;
  } // Most likely a plain Object


  return true;
}
function deepCloneStateWithoutComputed(source) {
  function recursiveClone(current) {
    const next = Object.keys(current).reduce((acc, key) => {
      if (Object.getOwnPropertyDescriptor(current, key).get == null) {
        acc[key] = current[key];
      }

      return acc;
    }, {});
    Object.keys(next).forEach(key => {
      if (isPlainObject(next[key])) {
        next[key] = recursiveClone(next[key]);
      }
    });
    return next;
  }

  return recursiveClone(source);
}
function isPromise(x) {
  return x != null && typeof x === 'object' && typeof x.then === 'function';
}
function get(path, target) {
  return path.reduce((acc, cur) => {
    return isPlainObject(acc) ? acc[cur] : undefined;
  }, target);
}
function newify(currentPath, currentState, finalValue) {
  if (currentPath.length === 0) {
    return finalValue;
  }

  const newState = _extends({}, currentState);

  const key = currentPath[0];

  if (currentPath.length === 1) {
    newState[key] = finalValue;
  } else {
    newState[key] = newify(currentPath.slice(1), newState[key], finalValue);
  }

  return newState;
}
function set(path, target, value) {
  if (path.length === 0) {
    if (typeof value === 'object') {
      Object.keys(target).forEach(key => {
        delete target[key];
      });
      Object.keys(value).forEach(key => {
        target[key] = value[key];
      });
    }

    return;
  }

  path.reduce((acc, cur, idx) => {
    if (idx + 1 === path.length) {
      acc[cur] = value;
    } else {
      acc[cur] = acc[cur] || {};
    }

    return acc[cur];
  }, target);
}
function createSimpleProduce(disableImmer = false) {
  return function simpleProduce(path, state, fn) {
    if (disableImmer) {
      const current = get(path, state);
      const next = fn(current);

      if (current !== next) {
        return newify(path, state, next);
      }

      return state;
    }

    if (path.length === 0) {
      const draft = easyPeasyImmer.createDraft(state);
      const result = fn(draft);

      if (result) {
        return isDraft(result) ? easyPeasyImmer.finishDraft(result) : result;
      }

      return easyPeasyImmer.finishDraft(draft);
    }

    const parentPath = path.slice(0, path.length - 1);
    const draft = easyPeasyImmer.createDraft(state);
    const parent = get(parentPath, state);
    const current = get(path, draft);
    const result = fn(current);

    if (result) {
      parent[path[path.length - 1]] = result;
    }

    return easyPeasyImmer.finishDraft(draft);
  };
}

const pReduce = (iterable, reducer, initialValue) => new Promise((resolve, reject) => {
  const iterator = iterable[Symbol.iterator]();
  let index = 0;

  const next = async total => {
    const element = iterator.next();

    if (element.done) {
      resolve(total);
      return;
    }

    try {
      const value = await Promise.all([total, element.value]); // eslint-disable-next-line no-plusplus

      next(reducer(value[0], value[1], index++));
    } catch (error) {
      reject(error);
    }
  };

  next(initialValue);
});

const pSeries = async tasks => {
  const results = [];
  await pReduce(tasks, async (_, task) => {
    const value = await task();
    results.push(value);
  });
  return results;
};
function areInputsEqual(newInputs, lastInputs) {
  if (newInputs.length !== lastInputs.length) {
    return false;
  }

  for (let i = 0; i < newInputs.length; i += 1) {
    if (newInputs[i] !== lastInputs[i]) {
      return false;
    }
  }

  return true;
}
function memoizeOne(resultFn) {
  let lastArgs = [];
  let lastResult;
  let calledOnce = false;
  return function memoized(...args) {
    if (calledOnce && areInputsEqual(args, lastArgs)) {
      return lastResult;
    }

    lastResult = resultFn(...args);
    calledOnce = true;
    lastArgs = args;
    return lastResult;
  };
}
function useMemoOne( // getResult changes on every call,
getResult, // the inputs array changes on every call
inputs) {
  // using useState to generate initial value as it is lazy
  const initial = React.useState(() => ({
    inputs,
    result: getResult()
  }))[0];
  const committed = React.useRef(initial); // persist any uncommitted changes after they have been committed

  const isInputMatch = Boolean(inputs && committed.current.inputs && areInputsEqual(inputs, committed.current.inputs)); // create a new cache if required

  const cache = isInputMatch ? committed.current : {
    inputs,
    result: getResult()
  }; // commit the cache

  React.useEffect(() => {
    committed.current = cache;
  }, [cache]);
  return cache.result;
}

function createReducer(disableImmer, _actionReducersDict, _customReducers, _computedProperties) {
  const simpleProduce = createSimpleProduce(disableImmer);

  const runActionReducerAtPath = (state, action, actionReducer, path) => {
    return simpleProduce(path, state, draft => actionReducer(draft, action.payload));
  };

  const reducerForActions = (state, action) => {
    const actionReducer = _actionReducersDict[action.type];

    if (actionReducer) {
      return runActionReducerAtPath(state, action, actionReducer, actionReducer.definition.meta.parent);
    }

    return state;
  };

  const reducerForCustomReducers = (state, action) => {
    return _customReducers.reduce((acc, {
      parentPath,
      key,
      reducer: red
    }) => {
      return simpleProduce(parentPath, acc, draft => {
        draft[key] = red(draft[key], action);
        return draft;
      });
    }, state);
  };

  const rootReducer = (state, action) => {
    const stateAfterActions = reducerForActions(state, action);
    const next = _customReducers.length > 0 ? reducerForCustomReducers(stateAfterActions, action) : stateAfterActions;

    if (state !== next) {
      _computedProperties.forEach(({
        parentPath,
        bindComputedProperty
      }) => {
        const parentState = get(parentPath, next);
        if (parentState != null) bindComputedProperty(parentState, next);
      });
    }

    return next;
  };

  return rootReducer;
}

const noopStorage = {
  getItem: () => undefined,
  setItem: () => undefined,
  removeItem: () => undefined
};

const getBrowerStorage = storageName => {
  let storageCache;
  return () => {
    if (!storageCache) {
      try {
        if (typeof window !== 'undefined' && typeof window[storageName] !== 'undefined') {
          storageCache = window[storageName];
        }
      } catch (_) {// swallow the failure
      }

      if (!storageCache) {
        storageCache = noopStorage;
      }
    }

    return storageCache;
  };
};

const localStorage = getBrowerStorage('localStorage');
const sessionStorage = getBrowerStorage('sessionStorage');

function createStorageWrapper(storage, transformers = []) {
  if (storage == null) {
    storage = sessionStorage();
  }

  if (typeof storage === 'string') {
    if (storage === 'localStorage') {
      storage = localStorage();
    } else if (storage === 'sessionStorage') {
      storage = sessionStorage();
    } else {
      if (process.env.NODE_ENV === 'development') {
        console.warn(`Invalid storage provider`);
      }

      storage = noopStorage;
    }
  }

  const outTransformers = [...transformers].reverse();

  const serialize = data => {
    if (transformers.length > 0 && data != null && typeof data === 'object') {
      Object.keys(data).forEach(key => {
        data[key] = transformers.reduce((acc, cur) => {
          return cur.in(acc, key);
        }, data[key]);
      });
    }

    return storage === localStorage() || storage === sessionStorage() ? JSON.stringify({
      data
    }) : data;
  };

  const deserialize = data => {
    const result = storage === localStorage() || storage === sessionStorage() ? JSON.parse(data).data : data;

    if (outTransformers.length > 0 && result != null && typeof result === 'object') {
      Object.keys(result).forEach(key => {
        result[key] = outTransformers.reduce((acc, cur) => {
          return cur.out(acc, key);
        }, result[key]);
      });
    }

    return result;
  };

  const isAsync = isPromise(storage.getItem('_'));
  return {
    isAsync,
    getItem: key => {
      if (isAsync) {
        return storage.getItem(key).then(wrapped => {
          return wrapped != null ? deserialize(wrapped) : undefined;
        });
      }

      const wrapped = storage.getItem(key);
      return wrapped != null ? deserialize(wrapped) : undefined;
    },
    setItem: (key, data) => {
      return storage.setItem(key, serialize(data));
    },
    removeItem: key => {
      return storage.removeItem(key);
    }
  };
}

function extractPersistConfig(path, persistDefinition = {}) {
  return {
    path,
    config: {
      allow: persistDefinition.allow || [],
      deny: persistDefinition.deny || [],
      mergeStrategy: persistDefinition.mergeStrategy || 'mergeDeep',
      storage: createStorageWrapper(persistDefinition.storage, persistDefinition.transformers)
    }
  };
}

function resolvePersistTargets(target, allow, deny) {
  let targets = Object.keys(target);

  if (allow.length > 0) {
    targets = targets.reduce((acc, cur) => {
      if (allow.findIndex(x => x === cur) !== -1) {
        return [...acc, cur];
      }

      return acc;
    }, []);
  }

  if (deny.length > 0) {
    targets = targets.reduce((acc, cur) => {
      if (deny.findIndex(x => x === cur) !== -1) {
        return acc;
      }

      return [...acc, cur];
    }, []);
  }

  return targets;
}

function createPersistenceClearer(persistKey, references) {
  return () => {
    if (references.internals._persistenceConfig.length === 0) {
      return Promise.resolve();
    }

    return pSeries(references.internals._persistenceConfig.map(({
      path,
      config
    }) => () => Promise.resolve(config.storage.removeItem(persistKey(path)))));
  };
}

function createPersistor(persistKey, references) {
  let persistPromise = Promise.resolve();
  let isPersisting = false;
  let nextPersistOperation;
  const timingMethod = typeof window === 'undefined' ? fn => fn() : window.requestIdleCallback != null ? window.requestIdleCallback : window.requestAnimationFrame;

  const persist = nextState => {
    if (references.internals._persistenceConfig.length === 0) {
      return;
    }

    const operation = () => {
      isPersisting = true;
      persistPromise = new Promise(resolve => {
        timingMethod(() => {
          pSeries(references.internals._persistenceConfig.map(({
            path,
            config
          }) => () => {
            const {
              storage,
              allow,
              deny
            } = config;
            const persistRootState = deepCloneStateWithoutComputed(get(path, nextState));
            const persistTargets = resolvePersistTargets(persistRootState, allow, deny);
            const stateToPersist = {};
            persistTargets.map(key => {
              const targetPath = [...path, key];
              const rawValue = get(targetPath, nextState);
              const value = isPlainObject(rawValue) ? deepCloneStateWithoutComputed(rawValue) : rawValue;
              stateToPersist[key] = value;
            });
            return Promise.resolve(storage.setItem(persistKey(path), stateToPersist));
          })).finally(() => {
            isPersisting = false;

            if (nextPersistOperation) {
              const next = nextPersistOperation;
              nextPersistOperation = null;
              next();
            } else {
              resolve();
            }
          });
        });
      });
    };

    if (isPersisting) {
      nextPersistOperation = operation;
    } else {
      operation();
    }
  };

  return {
    persist,
    clear: createPersistenceClearer(persistKey, references),
    flush: async () => {
      if (nextPersistOperation) {
        nextPersistOperation();
      }

      await persistPromise;
    }
  };
}
function createPersistMiddleware(persistor, references) {
  return ({
    getState
  }) => next => action => {
    const state = next(action);

    if (action && action.type !== '@action.ePRS' && references.internals._persistenceConfig.length > 0) {
      persistor.persist(getState());
    }

    return state;
  };
}
function rehydrateStateFromPersistIfNeeded(persistKey, replaceState, references, root) {
  if (references.internals._persistenceConfig.length === 0) {
    return Promise.resolve();
  }

  return pSeries(references.internals._persistenceConfig.map(persistInstance => () => {
    const {
      path,
      config
    } = persistInstance;
    const {
      mergeStrategy,
      storage
    } = config;

    if (root && (path.length < 1 || path[0] !== root)) {
      return Promise.resolve();
    }

    const state = references.internals._defaultState;

    const hasDataModelChanged = (dataModel, rehydratingModelData) => dataModel != null && rehydratingModelData != null && (typeof dataModel !== typeof rehydratingModelData || Array.isArray(dataModel) && !Array.isArray(rehydratingModelData));

    const applyRehydrationStrategy = (originalState, persistedState) => {
      if (mergeStrategy === 'overwrite') {
        set(path, originalState, persistedState);
      } else if (mergeStrategy === 'mergeShallow') {
        const targetState = get(path, originalState);
        Object.keys(persistedState).forEach(key => {
          if (hasDataModelChanged(targetState[key], persistedState[key])) ; else {
            targetState[key] = persistedState[key];
          }
        });
      } else if (mergeStrategy === 'mergeDeep') {
        const targetState = get(path, originalState);

        const setAt = (currentTargetState, currentPersistedState) => {
          Object.keys(currentPersistedState).forEach(key => {
            if (hasDataModelChanged(currentTargetState[key], currentPersistedState[key])) ; else if (isPlainObject(currentPersistedState[key])) {
              currentTargetState[key] = currentTargetState[key] || {};
              setAt(currentTargetState[key], currentPersistedState[key]);
            } else {
              currentTargetState[key] = currentPersistedState[key];
            }
          });
        };

        setAt(targetState, persistedState);
      }
    };

    const rehydate = persistedState => {
      if (persistedState != null) {
        applyRehydrationStrategy(state, persistedState);
      }

      replaceState(state);
    };

    const getItemResult = storage.getItem(persistKey(path));

    if (isPromise(getItemResult)) {
      return getItemResult.then(rehydate);
    }

    return Promise.resolve(rehydate(getItemResult));
  }));
}

function createActionCreator(definition, references) {
  const actionCreator = payload => {
    const action = {
      type: definition.meta.type,
      payload
    };

    if (definition[actionOnSymbol] && definition.meta.resolvedTargets) {
      payload.resolvedTargets = [...definition.meta.resolvedTargets];
    }

    return references.dispatch(action);
  }; // We bind the types to the creator for easy reference by consumers


  actionCreator.type = definition.meta.type;
  return actionCreator;
}

function createThunkHandler(definition, references, injections, _actionCreators) {
  return (payload, fail) => {
    const helpers = {
      dispatch: references.dispatch,
      fail,
      getState: () => get(definition.meta.parent, references.getState()),
      getStoreActions: () => _actionCreators,
      getStoreState: references.getState,
      injections,
      meta: {
        key: definition.meta.actionName,
        parent: definition.meta.parent,
        path: definition.meta.path
      }
    };

    if (definition[thunkOnSymbol] && definition.meta.resolvedTargets) {
      payload.resolvedTargets = [...definition.meta.resolvedTargets];
    }

    return definition.fn(get(definition.meta.parent, _actionCreators), payload, helpers);
  };
}

const logThunkEventListenerError = (type, err) => {
  // eslint-disable-next-line no-console
  console.log(`Error in ${type}`); // eslint-disable-next-line no-console

  console.log(err);
};

const handleEventDispatchErrors = (type, dispatcher) => (...args) => {
  try {
    const result = dispatcher(...args);

    if (isPromise(result)) {
      result.catch(err => {
        logThunkEventListenerError(type, err);
      });
    }
  } catch (err) {
    logThunkEventListenerError(type, err);
  }
};

function createThunkActionsCreator(definition, references) {
  const actionCreator = payload => {
    const dispatchStart = handleEventDispatchErrors(definition.meta.startType, () => references.dispatch({
      type: definition.meta.startType,
      payload
    }));
    const dispatchFail = handleEventDispatchErrors(definition.meta.failType, err => references.dispatch({
      type: definition.meta.failType,
      payload,
      error: err
    }));
    const dispatchSuccess = handleEventDispatchErrors(definition.meta.successType, result => references.dispatch({
      type: definition.meta.successType,
      payload,
      result
    }));
    dispatchStart();
    let failure = null;

    const fail = _failure => {
      failure = _failure;
    };

    const result = references.dispatch(() => definition.thunkHandler(payload, fail));

    if (isPromise(result)) {
      return result.then(resolved => {
        if (failure) {
          dispatchFail(failure);
        } else {
          dispatchSuccess(resolved);
        }

        return resolved;
      });
    }

    if (failure) {
      dispatchFail(failure);
    } else {
      dispatchSuccess(result);
    }

    return result;
  };

  actionCreator.type = definition.meta.type;
  actionCreator.successType = definition.meta.successType;
  actionCreator.failType = definition.meta.failType;
  actionCreator.startType = definition.meta.startType;
  return actionCreator;
}

function createListenerMiddleware(references) {
  return () => next => action => {
    const result = next(action);

    if (action && references.internals._listenerActionMap[action.type] && references.internals._listenerActionMap[action.type].length > 0) {
      const sourceAction = references.internals._actionCreatorDict[action.type];

      references.internals._listenerActionMap[action.type].forEach(actionCreator => {
        actionCreator({
          type: sourceAction ? sourceAction.definition.meta.type : action.type,
          payload: action.payload,
          error: action.error,
          result: action.result
        });
      });
    }

    return result;
  };
}
function bindListenerDefinitions(listenerDefinitions, _actionCreators, _actionCreatorDict, _listenerActionMap) {
  listenerDefinitions.forEach(definition => {
    const targets = definition.targetResolver(get(definition.meta.parent, _actionCreators), _actionCreators);
    const targetTypes = (Array.isArray(targets) ? targets : [targets]).reduce((acc, target) => {
      if (typeof target === 'function' && target.definition.meta.type && _actionCreatorDict[target.definition.meta.type]) {
        if (target.definition.meta.successType) {
          acc.push(target.definition.meta.successType); // acc.push(target.definition.meta.failType);
        } else {
          acc.push(target.definition.meta.type);
        }
      } else if (typeof target === 'string') {
        acc.push(target);
      }

      return acc;
    }, []);
    definition.meta.resolvedTargets = targetTypes;
    targetTypes.forEach(targetType => {
      const listenerReg = _listenerActionMap[targetType] || [];
      listenerReg.push(_actionCreatorDict[definition.meta.type]);
      _listenerActionMap[targetType] = listenerReg;
    });
  });
}

function createComputedPropertyBinder(parentPath, key, definition) {
  const memoisedResultFn = memoizeOne(definition.fn);
  return function createComputedProperty(parentState, storeState) {
    Object.defineProperty(parentState, key, {
      configurable: true,
      enumerable: true,
      get: () => {
        const state = get(parentPath, storeState);
        const inputs = definition.stateResolvers.map(resolver => resolver(state, storeState));
        return memoisedResultFn(...inputs);
      }
    });
  };
}
function createComputedPropertiesMiddleware(references) {
  return store => next => action => {
    references.internals._computedState.currentState = store.getState();
    references.internals._computedState.isInReducer = true;
    return next(action);
  };
}

function createEffectsMiddleware(references) {
  return store => next => action => {
    if (references.internals._effects.length === 0) {
      return next(action);
    }

    const prevState = store.getState();
    const result = next(action);
    const nextState = store.getState();

    references.internals._effects.forEach(definition => {
      const prevLocal = get(definition.meta.parent, prevState);
      const nextLocal = get(definition.meta.parent, nextState);

      if (prevLocal !== nextLocal) {
        const prevDependencies = definition.dependencyResolvers.map(resolver => resolver(prevLocal));
        const nextDependencies = definition.dependencyResolvers.map(resolver => resolver(nextLocal));
        const hasChanged = prevDependencies.some((dependency, idx) => {
          return dependency !== nextDependencies[idx];
        });

        if (hasChanged) {
          definition.actionCreator(prevDependencies, nextDependencies, action);
        }
      }
    });

    return result;
  };
}

const logEffectError = err => {
  // As users can't get a handle on effects we need to report the error
  // eslint-disable-next-line no-console
  console.log(err);
};

function createEffectHandler(definition, references, injections, _actionCreators) {
  const actions = get(definition.meta.parent, _actionCreators);
  let dispose;
  return change => {
    const helpers = {
      dispatch: references.dispatch,
      getState: () => get(definition.meta.parent, references.getState()),
      getStoreActions: () => _actionCreators,
      getStoreState: references.getState,
      injections,
      meta: {
        key: definition.meta.actionName,
        parent: definition.meta.parent,
        path: definition.meta.path
      }
    };

    if (dispose !== undefined) {
      const disposeResult = dispose();
      dispose = undefined;

      if (isPromise(disposeResult)) {
        disposeResult.catch(logEffectError);
      }
    }

    const effectResult = definition.fn(actions, change, helpers);

    if (isPromise(effectResult)) {
      return effectResult.then(resolved => {
        if (typeof resolved === 'function') {
          if (process.env.NODE_ENV !== 'production') {
            // Dispose functions are not allowed to be resolved asynchronously.
            // Doing so would provide inconsistent behaviour around their execution.
            // eslint-disable-next-line no-console
            console.warn('[easy-peasy] Effect is asynchronously resolving a dispose fn.');
          }
        }
      });
    }

    if (typeof effectResult === 'function') {
      dispose = effectResult;
    }

    return undefined;
  };
}

const logEffectEventListenerError = (type, err) => {
  // eslint-disable-next-line no-console
  console.log(`Error in ${type}`); // eslint-disable-next-line no-console

  console.log(err);
};

const handleEventDispatchErrors$1 = (type, dispatcher) => (...args) => {
  try {
    const result = dispatcher(...args);

    if (isPromise(result)) {
      result.catch(err => {
        logEffectEventListenerError(type, err);
      });
    }
  } catch (err) {
    logEffectEventListenerError(type, err);
  }
};

function createEffectActionsCreator(definition, references, effectHandler) {
  const actionCreator = (previousDependencies, nextDependencies, action) => {
    const change = {
      prev: previousDependencies,
      current: nextDependencies,
      action
    };
    const dispatchStart = handleEventDispatchErrors$1(definition.meta.startType, () => references.dispatch({
      type: definition.meta.startType,
      change
    }));
    const dispatchSuccess = handleEventDispatchErrors$1(definition.meta.successType, () => references.dispatch({
      type: definition.meta.successType,
      change
    }));
    dispatchStart();

    try {
      const result = references.dispatch(() => effectHandler(change));

      if (isPromise(result)) {
        return result.then(resolved => {
          dispatchSuccess(resolved);
          return resolved;
        }, logEffectError);
      }

      dispatchSuccess(result);
      return result;
    } catch (err) {
      logEffectError(err);
    }
  };

  actionCreator.type = definition.meta.type;
  actionCreator.startType = definition.meta.startType;
  actionCreator.successType = definition.meta.successType;
  actionCreator.failType = definition.meta.failType;
  return actionCreator;
}

function extractDataFromModel(model, initialState, injections, references) {
  const _defaultState = initialState;
  const _actionCreatorDict = {};
  const _actionCreators = {};
  const _actionReducersDict = {};
  const actionThunks = {};
  const _computedProperties = [];
  const _customReducers = [];
  const _effects = [];
  const _listenerActionCreators = {};
  const _listenerActionMap = {};
  const listenerDefinitions = [];
  let _persistenceConfig = [];
  const _computedState = {
    isInReducer: false,
    currentState: _defaultState
  };

  const recursiveExtractFromModel = (current, parentPath) => Object.keys(current).forEach(key => {
    const value = current[key];
    const path = [...parentPath, key];
    const meta = {
      parent: parentPath,
      path,
      key
    };

    const handleValueAsState = () => {
      const initialParentRef = get(parentPath, initialState);

      if (initialParentRef && key in initialParentRef) {
        set(path, _defaultState, initialParentRef[key]);
      } else {
        set(path, _defaultState, value);
      }
    };

    if (key === persistSymbol) {
      _persistenceConfig.push(extractPersistConfig(parentPath, value));

      return;
    }

    if (value != null && typeof value === 'object') {
      if (value[actionSymbol] || value[actionOnSymbol]) {
        const definition = _extends({}, value); // Determine the category of the action


        const category = definition[actionSymbol] ? '@action' : '@actionOn'; // Establish the meta data describing the action

        definition.meta = {
          actionName: meta.key,
          category,
          type: `${category}.${meta.path.join('.')}`,
          parent: meta.parent,
          path: meta.path
        }; // Create the "action creator" function

        definition.actionCreator = createActionCreator(definition, references); // Create a bidirectional relationship of the definition/actionCreator

        definition.actionCreator.definition = definition; // Create a bidirectional relationship of the definition/reducer

        definition.fn.definition = definition; // Add the action creator to lookup map

        _actionCreatorDict[definition.meta.type] = definition.actionCreator; // Add the reducer to lookup map

        _actionReducersDict[definition.meta.type] = definition.fn; // We don't want to expose the internal action to consumers

        if (meta.key !== 'ePRS') {
          // Set the action creator in the "actions" object tree for
          // either the listeners object tree, or the standard actions/thunks
          // object tree
          if (definition[actionOnSymbol]) {
            listenerDefinitions.push(definition);
            set(path, _listenerActionCreators, definition.actionCreator);
          } else {
            set(path, _actionCreators, definition.actionCreator);
          }
        }
      } else if (value[thunkSymbol] || value[thunkOnSymbol]) {
        const definition = _extends({}, value); // Determine the category of the thunk


        const category = definition[thunkSymbol] ? '@thunk' : '@thunkOn'; // Establish the meta data describing the thunk

        const type = `${category}.${meta.path.join('.')}`;
        definition.meta = {
          actionName: meta.key,
          parent: meta.parent,
          path: meta.path,
          type,
          startType: `${type}(start)`,
          successType: `${type}(success)`,
          failType: `${type}(fail)`
        }; // Create the function that will handle, i.e. be executed, when
        // the thunk action is created/dispatched

        definition.thunkHandler = createThunkHandler(definition, references, injections, _actionCreators); // Register the thunk handler

        set(path, actionThunks, definition.thunkHandler); // Create the "action creator" function

        definition.actionCreator = createThunkActionsCreator(definition, references); // Create a bidirectional relationship of the definition/actionCreator

        definition.actionCreator.definition = definition; // Register the action creator within the lookup map

        _actionCreatorDict[definition.meta.type] = definition.actionCreator; // Set the action creator in the "actions" object tree for
        // either the listeners object tree, or the standard actions/thunks
        // object tree

        if (definition[thunkOnSymbol]) {
          listenerDefinitions.push(definition);
          set(path, _listenerActionCreators, definition.actionCreator);
        } else {
          set(path, _actionCreators, definition.actionCreator);
        }
      } else if (value[computedSymbol]) {
        const parent = get(parentPath, _defaultState);
        const bindComputedProperty = createComputedPropertyBinder(parentPath, key, value);
        bindComputedProperty(parent, _defaultState);

        _computedProperties.push({
          key,
          parentPath,
          bindComputedProperty
        });
      } else if (value[reducerSymbol]) {
        _customReducers.push({
          key,
          parentPath,
          reducer: value.fn
        });
      } else if (value[effectOnSymbol]) {
        const definition = _extends({}, value); // Establish the meta data describing the effect


        const type = `@effectOn.${meta.path.join('.')}`;
        definition.meta = {
          type,
          actionName: meta.key,
          parent: meta.parent,
          path: meta.path,
          startType: `${type}(start)`,
          successType: `${type}(success)`,
          failType: `${type}(fail)`
        };
        const effectHandler = createEffectHandler(definition, references, injections, _actionCreators);
        const actionCreator = createEffectActionsCreator(definition, references, effectHandler);
        definition.actionCreator = actionCreator;

        _effects.push(definition);
      } else if (isPlainObject(value)) {
        const existing = get(path, _defaultState);

        if (existing == null) {
          set(path, _defaultState, {});
        }

        recursiveExtractFromModel(value, path);
      } else {
        handleValueAsState();
      }
    } else {
      handleValueAsState();
    }
  });

  _persistenceConfig = _persistenceConfig.sort((a, b) => {
    const aPath = a.path.join('.');
    const bPath = b.path.join('.');

    if (aPath < bPath) {
      return -1;
    }

    if (aPath > bPath) {
      return 1;
    }

    return 0;
  });
  recursiveExtractFromModel(model, []);
  bindListenerDefinitions(listenerDefinitions, _actionCreators, _actionCreatorDict, _listenerActionMap);
  return {
    _actionCreatorDict,
    _actionCreators,
    _actionReducersDict,
    _computedProperties,
    _customReducers,
    _computedState,
    _defaultState,
    _effects,
    _listenerActionCreators,
    _listenerActionMap,
    _persistenceConfig
  };
}

function createStoreInternals({
  disableImmer,
  initialState,
  injections,
  model,
  reducerEnhancer,
  references
}) {
  const {
    _actionCreatorDict,
    _actionCreators,
    _actionReducersDict,
    _computedState,
    _computedProperties,
    _customReducers,
    _defaultState,
    _effects,
    _listenerActionCreators,
    _listenerActionMap,
    _persistenceConfig
  } = extractDataFromModel(model, initialState, injections, references);
  const rootReducer = createReducer(disableImmer, _actionReducersDict, _customReducers, _computedProperties);
  return {
    _actionCreatorDict,
    _actionCreators,
    _computedProperties,
    _computedState,
    _defaultState,
    _effects,
    _listenerActionCreators,
    _listenerActionMap,
    _persistenceConfig,
    reducer: reducerEnhancer(rootReducer)
  };
}

function createStore(model, options = {}) {
  const modelClone = deepCloneStateWithoutComputed(model);
  const {
    compose: compose$1,
    devTools = process.env.NODE_ENV !== 'production',
    disableImmer = false,
    enhancers = [],
    initialState = {},
    injections = {},
    middleware = [],
    mockActions = false,
    name: storeName = `EasyPeasyStore`,
    version = 0,
    reducerEnhancer = rootReducer => rootReducer
  } = options;

  const bindReplaceState = modelDef => {
    return _extends({}, modelDef, {
      ePRS: action((_, payload) => payload)
    });
  };

  const references = {};
  let modelDefinition = bindReplaceState(modelClone);
  let mockedActions = [];

  const persistKey = targetPath => `[${storeName}][${version}]${targetPath.length > 0 ? `[${targetPath.join('.')}]` : ''}`;

  const persistor = createPersistor(persistKey, references);
  const persistMiddleware = createPersistMiddleware(persistor, references);

  const replaceState = nextState => references.internals._actionCreatorDict['@action.ePRS'](nextState);

  const bindStoreInternals = (state = {}) => {
    references.internals = createStoreInternals({
      disableImmer,
      initialState: state,
      injections,
      model: modelDefinition,
      reducerEnhancer,
      references
    });
  };

  const mockActionsMiddleware = () => () => action => {
    if (action != null) {
      mockedActions.push(action);
    }

    return undefined;
  };

  const composeEnhancers = compose$1 || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__({
    name: storeName
  }) : compose);
  bindStoreInternals(initialState);
  const easyPeasyMiddleware = [createComputedPropertiesMiddleware(references), ...middleware, reduxThunk, createListenerMiddleware(references), createEffectsMiddleware(references), persistMiddleware];

  if (mockActions) {
    easyPeasyMiddleware.push(mockActionsMiddleware);
  }

  const store = createStore$1(references.internals.reducer, references.internals._defaultState, composeEnhancers(applyMiddleware(...easyPeasyMiddleware), ...enhancers));
  store.subscribe(() => {
    references.internals._computedState.isInReducer = false;
  });
  references.dispatch = store.dispatch;
  references.getState = store.getState;

  const bindActionCreators = () => {
    Object.keys(store.dispatch).forEach(actionsKey => {
      delete store.dispatch[actionsKey];
    });
    Object.keys(references.internals._actionCreators).forEach(key => {
      store.dispatch[key] = references.internals._actionCreators[key];
    });
  };

  bindActionCreators();

  const rebindStore = removeKey => {
    const currentState = store.getState();

    if (removeKey) {
      delete currentState[removeKey];
    }

    bindStoreInternals(currentState);
    store.replaceReducer(references.internals.reducer);
    replaceState(references.internals._defaultState);
    bindActionCreators();
  };

  const resolveRehydration = rehydrateStateFromPersistIfNeeded(persistKey, replaceState, references);
  return Object.assign(store, {
    addModel: (key, modelForKey) => {
      if (modelDefinition[key] && process.env.NODE_ENV !== 'production') {
        store.removeModel(key);
      }

      modelDefinition[key] = modelForKey;
      rebindStore(); // There may have been persisted state for a dynamic model. We should try
      // and rehydrate the specifc node

      const addModelRehydration = rehydrateStateFromPersistIfNeeded(persistKey, replaceState, references, key);
      return {
        resolveRehydration: () => addModelRehydration
      };
    },
    clearMockedActions: () => {
      mockedActions = [];
    },
    getActions: () => references.internals._actionCreators,
    getListeners: () => references.internals._listenerActionCreators,
    getMockedActions: () => [...mockedActions],
    persist: {
      clear: persistor.clear,
      flush: persistor.flush,
      resolveRehydration: () => resolveRehydration
    },
    reconfigure: newModel => {
      modelDefinition = bindReplaceState(newModel);
      rebindStore();
    },
    removeModel: key => {
      if (!modelDefinition[key]) {
        return;
      }

      delete modelDefinition[key];
      rebindStore(key);
    }
  });
}

function createContextStore(model, config = {}) {
  // We create a mutable injections reference to allow updating it
  const {
    injections: mutableInjections = {}
  } = config;
  const StoreContext = createContext();

  function Provider({
    children,
    runtimeModel,
    injections
  }) {
    // If the user provided injections we need to ensure our mutable ref
    // is up to date. We could consider doing a shallow compare here?
    if (injections != null) {
      const nextInjections = typeof injections === 'function' ? injections(mutableInjections) : injections;
      const nextKeys = Object.keys(nextInjections);
      const removeKeys = Object.keys(mutableInjections).filter(k => !nextKeys.includes(k));
      removeKeys.forEach(k => {
        delete mutableInjections[k];
      });
      Object.assign(mutableInjections, nextInjections);
    }

    const store = useMemoOne(() => createStore(typeof model === 'function' ? model(runtimeModel) : model, _extends({}, config, {
      originalInjections: mutableInjections
    })), []);
    return /*#__PURE__*/React.createElement(StoreContext.Provider, {
      value: store
    }, children);
  }

  function useStore() {
    return useContext(StoreContext);
  }

  return {
    Provider,
    useStore,
    useStoreState: createStoreStateHook(StoreContext),
    useStoreActions: createStoreActionsHook(StoreContext),
    useStoreDispatch: createStoreDispatchHook(StoreContext),
    useStoreRehydrated: createStoreRehydratedHook(StoreContext)
  };
}

/**
 * This file has been copied from redux-persist.
 * The intention being to support as much of the redux-persist API as possible.
 */
function createTransform(inbound, outbound, config = {}) {
  const whitelist = config.whitelist || null;
  const blacklist = config.blacklist || null;

  function whitelistBlacklistCheck(key) {
    if (whitelist && whitelist.indexOf(key) === -1) return true;
    if (blacklist && blacklist.indexOf(key) !== -1) return true;
    return false;
  }

  return {
    in: (data, key, fullState) => !whitelistBlacklistCheck(key) && inbound ? inbound(data, key, fullState) : data,
    out: (data, key, fullState) => !whitelistBlacklistCheck(key) && outbound ? outbound(data, key, fullState) : data
  };
}

function StoreProvider({
  children,
  store
}) {
  return /*#__PURE__*/React.createElement(StoreContext.Provider, {
    value: store
  }, children);
}

function useLocalStore(modelCreator, dependencies = [], configCreator) {
  const storeRef = useRef();
  const configRef = useRef();
  const store = useMemoOne(() => {
    const previousState = storeRef.current != null ? storeRef.current.getState() : undefined;
    const config = configCreator != null ? configCreator(previousState, configRef.current) : undefined;

    const _store = createStore(modelCreator(previousState), config);

    configRef.current = config;
    storeRef.current = _store;
    return _store;
  }, dependencies);
  const [currentState, setCurrentState] = useState(() => store.getState());
  useEffect(() => {
    return store.subscribe(() => {
      const nextState = store.getState();

      if (currentState !== nextState) {
        setCurrentState(nextState);
      }
    });
  }, [store]);
  return [currentState, store.getActions(), store];
}

export { StoreProvider, action, actionOn, computed, createContextStore, createStore, createStoreActionsHook, createStoreDispatchHook, createStoreRehydratedHook, createStoreStateHook, createTransform, createTypedHooks, debug, generic, memo, persist, reducer, thunk, thunkOn, unstable_effectOn, useLocalStore, useStore, useStoreActions, useStoreDispatch, useStoreRehydrated, useStoreState };
//# sourceMappingURL=index.js.map
